﻿using System.Collections.Generic;

namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  /// <summary>
  /// Finds the index of the top lit-type item in the top type vector.
  /// </summary>
  /// <typeparam name="T"></typeparam>
  [StorableClass]
  public abstract class VectorIndexOfExpression<T> : StatelessExpression {
    protected VectorIndexOfExpression() { }
    [StorableConstructor]
    protected VectorIndexOfExpression(bool deserializing) : base(deserializing) { }

    protected bool IsNoop(IPushStack<List<T>> vectorStack, IPushStack<T> literalStack) {
      return vectorStack.IsEmpty || literalStack.IsEmpty;
    }

    protected void Eval(
      IInternalPushInterpreter interpreter,
      IPushStack<List<T>> vectorStack,
      IPushStack<T> literalStack,
      bool canOverride = false) {
      var vector = vectorStack.Pop();
      var literal = canOverride ? literalStack.Top : literalStack.Pop();
      var result = vector.IndexOf(literal);

      if (canOverride) interpreter.IntegerStack.SetTop(result);
      else interpreter.IntegerStack.Push(result);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.IntegerVector, "INTEGER[].INDEXOF", StackTypes.Integer)]
  public class IntegerVectorIndexOfExpression : VectorIndexOfExpression<long> {
    public IntegerVectorIndexOfExpression() { }
    [StorableConstructor]
    protected IntegerVectorIndexOfExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.IntegerVectorStack, interpreter.IntegerStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.IntegerVectorStack, interpreter.IntegerStack, true);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.FloatVector, "FLOAT[].INDEXOF", StackTypes.Float | StackTypes.Integer)]
  public class FloatVectorIndexOfExpression : VectorIndexOfExpression<double> {
    public FloatVectorIndexOfExpression() { }
    [StorableConstructor]
    protected FloatVectorIndexOfExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.FloatVectorStack, interpreter.FloatStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.FloatVectorStack, interpreter.FloatStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.BooleanVector, "BOOLEAN[].INDEXOF", StackTypes.Boolean | StackTypes.Integer)]
  public class BooleanVectorIndexOfExpression : VectorIndexOfExpression<bool> {
    public BooleanVectorIndexOfExpression() { }
    [StorableConstructor]
    protected BooleanVectorIndexOfExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.BooleanVectorStack, interpreter.BooleanStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.BooleanVectorStack, interpreter.BooleanStack);
    }
  }

  [StorableClass]
  [PushExpression(StackTypes.StringVector, "STRING[].INDEXOF", StackTypes.String | StackTypes.String)]
  public class StringVectorIndexOfExpression : VectorIndexOfExpression<string> {
    public StringVectorIndexOfExpression() { }
    [StorableConstructor]
    protected StringVectorIndexOfExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.StringVectorStack, interpreter.StringStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.StringVectorStack, interpreter.StringStack);
    }
  }
}
