﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Parser {
  using System.Collections.Generic;
  using System.Globalization;
  using System.Linq;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Expressions;

  public static class PushParser {
    private const string openBrace = "(";

    private const string closeBrace = ")";

    private const char delimiter = ' ';

    private static readonly CultureInfo cultureInfo = CultureInfo.CreateSpecificCulture("en-US");

    private static readonly char[] symbolTrim = { '\r', '\n' };

    public static Expression Parse(string source, int startIndex = 0) {
      var symbols = source.Split(delimiter);

      int endIndex;
      return Parse(symbols, 0, out endIndex);
    }

    public static PushProgram ParseProgram(string source, int startIndex = 0) {
      var expression = Parse(source, startIndex);

      return expression as PushProgram ?? new PushProgram(new[] { expression });
    }

    private static Expression Parse(string[] symbols, int startIndex, out int endIndex) {
      var expressions = new List<Expression>();

      for (var i = startIndex; i < symbols.Length; i++) {
        var symbol = symbols[i].TrimEnd(symbolTrim);

        if (string.IsNullOrWhiteSpace(symbol)) continue;

        switch (symbol) {
          case openBrace:
            var subExpression = Parse(symbols, i + 1, out endIndex);
            expressions.Insert(0, subExpression);
            i = endIndex;
            continue;
          case closeBrace:
            endIndex = i;
            return new PushProgram(expressions);
        }

        // literal
        Expression expression;
        if (TryParseLiteral(symbol, out expression)) {
          expressions.Insert(0, expression);
          continue;
        }

        // expression
        if (ExpressionTable.TryGetStatelessExpression(symbol, out expression)
         || ExpressionTable.TryGetStatefulExpression(symbol, out expression)) {
          expressions.Insert(0, expression);
          continue;
        }

        // identifier - custom expression or named literals
        expressions.Insert(0, new NameDefineXExecExpression(symbol));
      }

      endIndex = symbols.Length - 1;

      switch (expressions.Count) {
        case 0:
          return PushProgram.Empty;
        case 1:
          return expressions[0];
        default:
          return new PushProgram(expressions);
      }
    }

    private static bool TryParseLiteral(string word, out Expression expression) {
      if (word.StartsWith("'") && word.EndsWith("'") && word.Length == 3) {
        expression = new CharPushExpression(word[1]);
        return true;
      }

      if (word.StartsWith("\"") && word.EndsWith("\"")) {
        expression = new StringPushExpression(word);
        return true;
      }

      // "[]" has no values and can therefer not parsed correctly as vector of a specific type
      if (word.StartsWith("[") && word.EndsWith("]") && word.Length > 3) {
        var vectorEntries = word
          .Substring(1, word.Length - 2)
          .Split(',')
          .Select(x => {
            Expression e;
            var success = TryParseLiteral(x.Trim(), out e);

            return new { success, e, type = e.GetType() };
          })
          .ToArray();

        if (vectorEntries.Length == 0) {
          expression = null;
          return false;
        }

        var vectorEntryType = vectorEntries.First().type;
        if (vectorEntries.Any(x => !x.success || x.type != vectorEntryType)) {
          expression = null;
          return false;
        }

        var result = vectorEntries
          .Select(x => x.e)
          .ToArray();

        // TODO
        //switch(vectorEntryType.GetHashCode())
        //{
        //  case typeof(IntegerPushExpression).GetHashCode(): return new IntegerVectorPushExpression(result)
        //}

        //expression = new StringPushExpression(word);
        //return true;
      }

      long longValue;
      if (long.TryParse(word, out longValue)) {
        expression = new IntegerPushExpression(longValue);
        return true;
      }

      double floatValue;
      if (double.TryParse(word, NumberStyles.Float, cultureInfo, out floatValue)) {
        expression = new FloatPushExpression(floatValue);
        return true;
      }

      bool booleanValue;
      if (bool.TryParse(word, out booleanValue)) {
        expression = new BooleanPushExpression(booleanValue);
        return true;
      }

      expression = null;
      return false;
    }
  }
}