﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Windows.Forms;

namespace HeuristicLab.Problems.ProgramSynthesis.Push.Views {
  using Configuration;
  using Core.Views;
  using Expressions;
  using Interpreter;
  using MainForm;
  using Problem;
  using Stack;

  public partial class PushDebuggerView : ItemView {
    private readonly IDictionary<StackTypes, ListBox> debugControlDict = new Dictionary<StackTypes, ListBox>();

    private PooledPushInterpreter interpreter;
    private PushInterpreterPool pool;

    private const string GROUP_BOX_TEXT_STRING_FORMAT = "{0}[{1}]";
    private const string EXEC_COUNT_LABEL_FORMAT = "{0}/{1} max. expressions evaluated";

    public event EventHandler<IPushInterpreter> OnReset;

    public PushDebuggerView() {
      InitializeComponent();
      InitEvents();
    }

    ~PushDebuggerView() {
      interpreter.Dispose();
    }

    public new PushSolution Content
    {
      get { return (PushSolution)base.Content; }
      set
      {
        base.Content = value;
      }
    }

    protected override void OnContentChanged() {
      if (Content == null) return;

      Name = "Push Debugger";
      pool = new PushInterpreterPool(Content.Config);

      if (interpreter != null) {
        interpreter.Dispose();
      }

      interpreter = pool.Create(Content.Random);

      InitDebugLists(Content.Config);
      ClearDebugLists();
      UpdateExecList();
      UpdateDebugLists();
      ResetDebugging();
    }

    private void UpdateExecCountLabel() {
      execCountLabel.Text = string.Format(
        EXEC_COUNT_LABEL_FORMAT,
        interpreter.ExecCounter,
        Content.Config.EvalPushLimit);
    }

    private void InitEvents() {
      runButton.Click += RunButtonClick;
      resetButton.Click += ResetButtonClick;
      stepButton.Click += StepButtonClick;
      simplifyButton.Click += SimplifyButtonClick;
    }

    private void RunButtonClick(object sender, EventArgs e) {
      if (interpreter == null)
        return;

      interpreter.Resume();
      UpdateStep();
    }

    private void StepButtonClick(object sender, EventArgs e) {
      if (interpreter == null || stepWidthBox.Value <= 0)
        return;

      var count = Math.Min(stepWidthBox.Value, interpreter.ExecStack.Count);

      for (var i = 0; i < count; i++) {
        SkipNoops();

        // run next none noop expression
        interpreter.Step();
      }

      // skip trailing noops so next expression is a none noop in debugger
      SkipNoops();

      stepWidthBox.Maximum = Math.Max(1, interpreter.ExecStack.Count);
      UpdateStep();
    }

    private void UpdateStep() {
      UpdateExecList();
      UpdateDebugLists();
      UpdateExecCountLabel();
      CheckIfButtonsCanBeEnabled();
    }

    private void SkipNoops() {
      // skip noops
      if (skipNoopsCheckBox.Checked) {
        while (interpreter.CanStep && interpreter.ExecStack.Top.IsNoop(interpreter)) {
          interpreter.Step();
        }
      }
    }

    private void CheckIfButtonsCanBeEnabled() {
      runButton.Enabled = interpreter != null && interpreter.CanStep;
      stepButton.Enabled = interpreter != null && interpreter.CanStep;
      stepWidthBox.Enabled = interpreter != null && interpreter.CanStep;
    }

    private void ResetButtonClick(object sender, EventArgs e) {
      ResetDebugging();
    }

    private void SimplifyButtonClick(object sender, EventArgs e) {
      var newContent = Content.Simplify();

      MainFormManager.MainForm.ShowContent(newContent, GetType());
    }

    public void ResetDebugging() {
      if (Content == null ||
          pool == null ||
          Content.Program == null)
        return;

      if (interpreter != null) {
        interpreter.Reset();
      }

      if (OnReset != null) {
        OnReset(this, interpreter);
      }

      interpreter.Run(Content.Program, true);

      stepWidthBox.Maximum = interpreter.ExecStack.Count;
      UpdateStep();
    }

    private void ClearDebugLists() {
      foreach (var list in debugControlDict.Values) {
        list.Items.Clear();
      }
    }

    private void UpdateExecList() {
      execList.Items.Clear();
      var expressions = interpreter.ExecStack.AsEnumerable()
        .Reverse()
        .Select(e => e.StringRepresentation)
        .ToArray();

      execList.Items.AddRange(expressions);
      execGroupBox.Text = string.Format(GROUP_BOX_TEXT_STRING_FORMAT, Enum.GetName(typeof(StackTypes), StackTypes.Exec), interpreter.ExecStack.Count);
    }

    private void InitDebugLists(IReadOnlyPushConfiguration config) {
      debugControlDict.Clear();

      // 2 = ExecList + EmptyColumn which is required to fill empty space
      while (debugTableLayout.ColumnCount > 2) {
        debugTableLayout.Controls.RemoveAt(1);
        debugTableLayout.ColumnCount--;
      }

      foreach (StackTypes stackType in Enum.GetValues(typeof(StackTypes))) {
        if (stackType != StackTypes.Exec &&
            ExpressionTable.GetExpressionsByStackTypes(stackType).Intersect(config.EnabledExpressions).Any()) {
          var list = CreateDebugList(stackType);
          debugControlDict.Add(stackType, list);
        }
      }
    }

    private ListBox CreateDebugList(StackTypes type) {
      var groupBox = new GroupBox {
        Anchor = AnchorStyles.Bottom | AnchorStyles.Right | AnchorStyles.Left | AnchorStyles.Top,
        AutoSize = true,
        AutoSizeMode = AutoSizeMode.GrowAndShrink,
        Text = Enum.GetName(typeof(StackTypes), type)
      };

      var list = new ListBox {
        Dock = DockStyle.Fill
      };

      // align numbers right
      var stackEntryType = type.GetStackEntryType();
      if (stackEntryType == typeof(double) ||
          stackEntryType == typeof(long)) {
        list.DrawMode = DrawMode.OwnerDrawFixed;
        list.DrawItem += (sender, e) => {
          if (e.Index <= -1) return;
          var item = list.Items[e.Index];

          e.DrawBackground();
          e.DrawFocusRectangle();

          var brush = new SolidBrush(e.ForeColor);
          var size = e.Graphics.MeasureString(item.ToString(), e.Font);

          e.Graphics.DrawString(
            item.ToString(),
            e.Font,
            brush,
            e.Bounds.Right - size.Width,
            e.Bounds.Top + (e.Bounds.Height / 2 - size.Height / 2));
        };
      }

      groupBox.Controls.Add(list);

      var columnWidth = stackEntryType == typeof(Expression) ? 250 : 150;

      debugTableLayout.ColumnCount++;
      debugTableLayout.ColumnStyles.Insert(1, new ColumnStyle(SizeType.Absolute, columnWidth));
      debugTableLayout.Controls.Add(groupBox);
      debugTableLayout.Controls.SetChildIndex(groupBox, 1);

      return list;
    }


    private void UpdateDebugLists() {
      ClearDebugLists();

      if (Content == null || interpreter == null)
        return;

      foreach (var pair in debugControlDict) {
        var name = Enum.GetName(typeof(StackTypes), pair.Key);
        var items = InterpreterStackStringifier.StringifyStack(interpreter, pair.Key).ToArray();

        pair.Value.Items.AddRange(items);
        ((GroupBox)pair.Value.Parent).Text = string.Format(GROUP_BOX_TEXT_STRING_FORMAT, name, pair.Value.Items.Count);
      }
    }

  }
}
